package log

import (
	"fmt"
	"log"
	"os"
	"path"
	"runtime"
	"strings"
)

type Log struct {
	module string
}

func New(module string) *Log {
	return &Log{
		module: module,
	}
}

var Default = New("common")

// Internal logging function
func (l *Log) _log(ln bool, level Level, v ...interface{}) {
	if level < logLevel {
		return
	}
	prefix := `[` + l.module
	if logDebugTrace {
		pc, file, line, _ := runtime.Caller(3)
		parts := strings.Split(runtime.FuncForPC(pc).Name(), `.`)
		prefix += fmt.Sprintf(`/%s(%s:%d)`, parts[len(parts)-1], path.Base(file), line)
	}
	prefix += `]`

	_level := levelsToNames[level]

	if !ln {
		prefix += ` `
		_level += ` `
	}
	vPrefix := append([]interface{}{prefix}, v...)
	vPrefixLevel := append([]interface{}{_level}, vPrefix...)

	// console logging
	if ln {
		log.Println(vPrefixLevel...)
	} else {
		log.Print(vPrefixLevel...)
	}

	// syslog logging
	// find which function to call for this level
	if f, ok := levelsToSyslog[level]; ok {
		if ln {
			_ = f(fmt.Sprintln(vPrefix...))
		} else {
			_ = f(fmt.Sprint(vPrefix...))
		}
	}
}

// Generic Logging functions
func (l *Log) log(level Level, v ...interface{}) {
	l._log(false, level, v...)
}
func (l *Log) logln(level Level, v ...interface{}) {
	l._log(true, level, v...)
}
func (l *Log) logf(level Level, format string, v ...interface{}) {
	l._log(false, level, fmt.Sprintf(format, v...))
}

// Fatal
func (l *Log) Fatal(v ...interface{}) {
	l.log(ALERT, v...)
	os.Exit(1)
}
func (l *Log) Fatalln(v ...interface{}) {
	l.logln(ALERT, v...)
	os.Exit(1)
}
func (l *Log) Fatalf(format string, v ...interface{}) {
	l.logf(ALERT, format, v...)
	os.Exit(1)
}

// Panic
func (l *Log) Panic(v ...interface{}) {
	l.log(CRITIC, v...)
	panic(fmt.Sprint(v...))
}
func (l *Log) Panicln(v ...interface{}) {
	l.logln(CRITIC, v...)
	panic(fmt.Sprint(v...))
}
func (l *Log) Panicf(format string, v ...interface{}) {
	l.logf(CRITIC, format, v...)
	panic(fmt.Sprintf(format, v...))
}

// Specific logging functions (by Levels)
func (l *Log) Debug(v ...interface{})                 { l.log(DEBUG, v...) }
func (l *Log) Debugln(v ...interface{})               { l.logln(DEBUG, v...) }
func (l *Log) Debugf(format string, v ...interface{}) { l.logf(DEBUG, format, v...) }

func (l *Log) Verbose(v ...interface{})                 { l.log(VERBOSE, v...) }
func (l *Log) Verboseln(v ...interface{})               { l.logln(VERBOSE, v...) }
func (l *Log) Verbosef(format string, v ...interface{}) { l.logf(VERBOSE, format, v...) }

func (l *Log) Info(v ...interface{})                 { l.log(INFO, v...) }
func (l *Log) Infoln(v ...interface{})               { l.logln(INFO, v...) }
func (l *Log) Infof(format string, v ...interface{}) { l.logf(INFO, format, v...) }

func (l *Log) Notice(v ...interface{})                 { l.log(NOTICE, v...) }
func (l *Log) Noticeln(v ...interface{})               { l.logln(NOTICE, v...) }
func (l *Log) Noticef(format string, v ...interface{}) { l.logf(NOTICE, format, v...) }

func (l *Log) Warning(v ...interface{})                 { l.log(WARNING, v...) }
func (l *Log) Warningln(v ...interface{})               { l.logln(WARNING, v...) }
func (l *Log) Warningf(format string, v ...interface{}) { l.logf(WARNING, format, v...) }
func (l *Log) Warn(v ...interface{})                    { l.log(WARNING, v...) }
func (l *Log) Warnln(v ...interface{})                  { l.logln(WARNING, v...) }
func (l *Log) Warnf(format string, v ...interface{})    { l.logf(WARNING, format, v...) }

func (l *Log) Error(v ...interface{})                 { l.log(ERROR, v...) }
func (l *Log) Errorln(v ...interface{})               { l.logln(ERROR, v...) }
func (l *Log) Errorf(format string, v ...interface{}) { l.logf(ERROR, format, v...) }
func (l *Log) Errf(format string, v ...interface{})   { l.logf(ERROR, format, v...) }
func (l *Log) Err(v ...interface{})                   { l.logln(ERROR, v...) }
func (l *Log) Errln(v ...interface{})                 { l.log(ERROR, v...) }

func (l *Log) Critic(v ...interface{})                 { l.log(CRITIC, v...) }
func (l *Log) Criticln(v ...interface{})               { l.logln(CRITIC, v...) }
func (l *Log) Criticf(format string, v ...interface{}) { l.logf(CRITIC, format, v...) }
func (l *Log) Crit(v ...interface{})                   { l.log(CRITIC, v...) }
func (l *Log) Critln(v ...interface{})                 { l.logln(CRITIC, v...) }
func (l *Log) Critf(format string, v ...interface{})   { l.logf(CRITIC, format, v...) }

func (l *Log) Alert(v ...interface{})                 { l.log(ALERT, v...) }
func (l *Log) Alertln(v ...interface{})               { l.logln(ALERT, v...) }
func (l *Log) Alertf(format string, v ...interface{}) { l.logf(ALERT, format, v...) }
