package healthcheck

import (
	"context"
	"encoding/base64"
	"flag"
	"fmt"
	"os"
	"reflect"
	"regexp"
	"strings"
	"time"

	"github.com/open-io/oio-exporter/common"
	_log "github.com/open-io/oio-exporter/common/log"
	"github.com/open-io/oio-exporter/daemon"
	"gopkg.in/yaml.v2"
)

var log = _log.New("healthcheck")

var confInterval = time.Duration(0)
var confNoLocal = false

func InitFlags(fs *flag.FlagSet) {
	fs.DurationVar(&confInterval, "healthcheck.interval", confInterval, "[healthcheck] interval between polls")
	fs.BoolVar(&confNoLocal, "healthcheck.no-local", confNoLocal, "[healthcheck] disable local healthchecks")
}

type healthcheckDaemonFactory struct {
	defaultHealthchecks string
	healthchecksPath    string

	modules         map[string]*module
	modulesVersions map[string]int
	healthchecks    map[string]healthcheckI
}

func NewFactory(defaultHealthchecks, healthchecksPath string) daemon.DaemonFactoryI {
	return &healthcheckDaemonFactory{
		defaultHealthchecks: defaultHealthchecks,
		healthchecksPath:    healthchecksPath,
		modulesVersions:     make(map[string]int),
	}
}

func (df *healthcheckDaemonFactory) Daemons(inventory *common.Inventory) map[string]interface{} {
	params := make(map[string]interface{})

	if confNoLocal && !inventory.IsRemote {
		return params
	}

	namespace := inventory.Namespace
	hostname := inventory.Hostname

	for serviceType, services := range inventory.Services {
		_, ok := df.healthchecks[serviceType]
		if !ok {
			continue
		}

		for _, service := range services {
			params[service.ID] = &healthcheckDaemonParams{
				Namespace:   namespace,
				Hostname:    hostname,
				Id:          service.ID,
				Ip:          service.IP,
				Port:        service.Port,
				ServiceType: serviceType,
				Version:     df.modulesVersions[serviceType],
			}
		}
	}

	return params
}

func (df *healthcheckDaemonFactory) Reload() error {
	defaultHealthchecks := df.defaultHealthchecks
	healthchecksPath := df.healthchecksPath
	healthchecks := make(map[string]healthcheckI)

	var hc struct {
		Modules []*module
	}

	if healthchecksPath == "" {
		data, err := base64.StdEncoding.DecodeString(defaultHealthchecks)
		if err != nil {
			return err
		}

		if err := yaml.Unmarshal(data, &hc); err != nil {
			return err
		}
		log.Info("Healthchecks definition loaded from defaults")
	} else {
		f, err := os.Open(healthchecksPath)
		if err != nil {
			return err
		}
		defer f.Close()

		if err := yaml.NewDecoder(f).Decode(&hc); err != nil {
			return err
		}
		log.Infof("Healthchecks definition loaded from %s", healthchecksPath)
	}

	modules := make(map[string]*module)
	for _, module := range hc.Modules {
		var err error

		if module.Prober == "none" {
			continue
		}

		moduleName := module.Name

		ok := true
		if module.Timeout, err = time.ParseDuration(module.TimeoutStr); err != nil {
			log.Warnf(`failed to parse 'timeout' for module %s: "%s"`, moduleName, err)
			ok = false
		}
		if module.ExpectStr != "" {
			if module.Expect, err = regexp.Compile(module.ExpectStr); err != nil {
				log.Warnf(`failed to parse 'expect' for module %s: "%s"`, moduleName, err)
				ok = false
			}
		}
		switch module.Prober {
		case "http":
			healthchecks[moduleName] = &HTTPCheck{module}
		case "tcp":
			healthchecks[moduleName] = &TCPCheck{module}
		default:
			log.Warnf(`unknown 'prober' for module %s: "%s"`, moduleName, module.Prober)
			ok = false
		}

		if !ok {
			log.Warnf("ignoring module %s", moduleName)
			continue
		}

		modules[moduleName] = module
		if prevModule, ok := df.modules[moduleName]; ok && !module.Equal(prevModule) {
			log.Infof("Module '%s' updated", moduleName)

			df.modulesVersions[module.Name]++
		}
	}

	df.modules = modules
	df.healthchecks = healthchecks

	log.Infof("Loaded %d healthchecks definitions", len(healthchecks))

	return nil
}

func (df *healthcheckDaemonFactory) New(paramsI interface{}) daemon.DaemonI {
	params := paramsI.(*healthcheckDaemonParams)

	return &healthcheckDaemon{
		healthcheck: df.healthchecks[params.ServiceType],

		namespace:   params.Namespace,
		hostname:    strings.Split(params.Hostname, ":")[0],
		id:          params.Id,
		ip:          params.Ip,
		port:        params.Port,
		serviceType: params.ServiceType,
	}
}

type healthcheckDaemonParams struct {
	Namespace   string
	Hostname    string
	Id          string
	Ip          string
	Port        int
	ServiceType string
	Version     int
}

type healthcheckDaemon struct {
	daemon.BaseDaemon

	healthcheck healthcheckI

	namespace   string
	hostname    string
	id          string
	ip          string
	port        int
	serviceType string
}

type module struct {
	Name        string            `yaml:"name"`
	Prober      string            `yaml:"prober"`
	TimeoutStr  string            `yaml:"timeout"`
	Send        string            `yaml:"send"`
	ExpectStr   string            `yaml:"expect"`
	Method      string            `yaml:"method"`
	Headers     map[string]string `yaml:"headers"`
	StatusCodes []int             `yaml:"status_codes"`
	Route       string            `yaml:"route"`
	Timeout     time.Duration     `yaml:"_disabled"`
	Expect      *regexp.Regexp    `yaml:"_disabled2"`
}

// To exclude Timeout & Expect
func (m1 *module) Equal(m2 *module) bool {
	return m2 != nil &&
		m1.Name == m2.Name &&
		m1.Prober == m2.Prober &&
		m1.TimeoutStr == m2.TimeoutStr &&
		m1.Send == m2.Send &&
		m1.ExpectStr == m2.ExpectStr &&
		m1.Method == m2.Method &&
		reflect.DeepEqual(m1.Headers, m2.Headers) &&
		reflect.DeepEqual(m1.StatusCodes, m2.StatusCodes) &&
		m1.Route == m2.Route
}

type healthcheckI interface {
	check(ctx context.Context, targetIP string, targetPort int) (bool, error)
}

func (d *healthcheckDaemon) Interval() time.Duration {
	return confInterval
}

func (d *healthcheckDaemon) Collect(ctx context.Context, m *common.MetricData) error {
	val := float64(0)
	passed, _ := d.healthcheck.check(ctx, d.ip, d.port)
	if passed {
		val = float64(1)
	}
	m.Set(&common.Metric{
		Name: "probe_success",
		Labels: [][]string{
			[]string{"namespace", d.namespace},
			[]string{"target", d.hostname},
			[]string{"service", d.id},
			[]string{"service_type", d.serviceType},
			[]string{"service_addr", fmt.Sprintf("%s:%d", d.ip, d.port)},
		},
		Value: val,
		TTL:   confInterval * 2,
	})

	return nil
}
